// +build functional

package cri_containerd

import (
	"bufio"
	"context"
	"io/ioutil"
	"os"
	"path/filepath"
	"strings"
	"testing"
	"time"

	runtime "k8s.io/kubernetes/pkg/kubelet/apis/cri/runtime/v1alpha2"
)

func Test_Container_Network_LCOW(t *testing.T) {
	pullRequiredLcowImages(t, []string{imageLcowK8sPause, imageLcowAlpine})

	// create a directory and log file
	dir, err := ioutil.TempDir("", "")
	if err != nil {
		t.Fatalf("failed creating temp dir: %v", err)
	}
	defer func() {
		if err := os.RemoveAll(dir); err != nil {
			t.Fatalf("failed deleting temp dir: %v", err)
		}
	}()
	log := filepath.Join(dir, "ping.txt")

	sandboxRequest := &runtime.RunPodSandboxRequest{
		Config: &runtime.PodSandboxConfig{
			Metadata: &runtime.PodSandboxMetadata{
				Name:      t.Name() + "-Sandbox",
				Namespace: testNamespace,
			},
		},
		RuntimeHandler: lcowRuntimeHandler,
	}

	client := newTestRuntimeClient(t)
	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	podID := runPodSandbox(t, client, ctx, sandboxRequest)
	defer removePodSandbox(t, client, ctx, podID)
	defer stopPodSandbox(t, client, ctx, podID)

	request := &runtime.CreateContainerRequest{
		PodSandboxId: podID,
		Config: &runtime.ContainerConfig{
			Metadata: &runtime.ContainerMetadata{
				Name: t.Name() + "-Container",
			},
			Image: &runtime.ImageSpec{
				Image: imageLcowAlpine,
			},
			Command: []string{
				"ping",
				"-q", // -q outputs ping stats only.
				"-c",
				"10",
				"google.com",
			},
			LogPath: log,
			Linux:   &runtime.LinuxContainerConfig{},
		},
		SandboxConfig: sandboxRequest.Config,
	}

	containerID := createContainer(t, client, ctx, request)
	defer removeContainer(t, client, ctx, containerID)

	startContainer(t, client, ctx, containerID)
	defer stopContainer(t, client, ctx, containerID)

	// wait a while for container to write to stdout
	time.Sleep(3 * time.Second)

	// open the log and test for any packet loss
	logFile, err := os.Open(log)
	if err != nil {
		t.Fatal(err)
	}
	defer logFile.Close()

	s := bufio.NewScanner(logFile)
	for s.Scan() {
		v := strings.Fields(s.Text())
		t.Logf("ping output: %v", v)

		if v != nil && v[len(v)-1] == "loss" && v[len(v)-3] != "0%" {
			t.Fatalf("expected 0%% packet loss, got %v packet loss", v[len(v)-3])
		}
	}
}

func Test_Container_Network_Hostname(t *testing.T) {
	type config struct {
		name string

		runtimeHandler string
		sandboxImage   string
		containerImage string
		cmd            []string
	}
	tests := []config{
		{
			name:           "WCOW_Process",
			runtimeHandler: wcowProcessRuntimeHandler,
			sandboxImage:   imageWindowsRS5Nanoserver,
			containerImage: imageWindowsRS5Nanoserver,
			cmd:            []string{"cmd", "/c", "ping", "-t", "127.0.0.1"},
		},
		{
			name:           "WCOW_Hypervisor",
			runtimeHandler: wcowHypervisorRuntimeHandler,
			sandboxImage:   imageWindowsRS5Nanoserver,
			containerImage: imageWindowsRS5Nanoserver,
			cmd:            []string{"cmd", "/c", "ping", "-t", "127.0.0.1"},
		},
		{
			name:           "LCOW",
			runtimeHandler: lcowRuntimeHandler,
			sandboxImage:   imageLcowK8sPause,
			containerImage: imageLcowAlpine,
			cmd:            []string{"top"},
		},
	}

	for _, test := range tests {
		t.Run(test.name, func(t *testing.T) {
			if test.runtimeHandler == lcowRuntimeHandler {
				pullRequiredLcowImages(t, []string{test.sandboxImage, test.containerImage})
			} else {
				pullRequiredImages(t, []string{test.sandboxImage, test.containerImage})
			}

			sandboxRequest := &runtime.RunPodSandboxRequest{
				Config: &runtime.PodSandboxConfig{
					Metadata: &runtime.PodSandboxMetadata{
						Name:      t.Name() + "-Sandbox",
						Namespace: testNamespace,
					},
					Hostname: "TestHost",
				},
				RuntimeHandler: test.runtimeHandler,
			}

			client := newTestRuntimeClient(t)
			ctx, cancel := context.WithCancel(context.Background())
			defer cancel()

			podID := runPodSandbox(t, client, ctx, sandboxRequest)
			defer removePodSandbox(t, client, ctx, podID)
			defer stopPodSandbox(t, client, ctx, podID)

			containerRequest := &runtime.CreateContainerRequest{
				PodSandboxId: podID,
				Config: &runtime.ContainerConfig{
					Metadata: &runtime.ContainerMetadata{
						Name: t.Name() + "-Container",
					},
					Image: &runtime.ImageSpec{
						Image: test.containerImage,
					},
					Command: test.cmd,
				},
				SandboxConfig: sandboxRequest.Config,
			}

			containerID := createContainer(t, client, ctx, containerRequest)
			defer removeContainer(t, client, ctx, containerID)

			startContainer(t, client, ctx, containerID)
			defer stopContainer(t, client, ctx, containerID)

			execResponse := execSync(t, client, ctx, &runtime.ExecSyncRequest{
				ContainerId: containerID,
				Cmd:         []string{"hostname"},
			})
			stdout := strings.Trim(string(execResponse.Stdout), " \r\n")
			if stdout != sandboxRequest.Config.Hostname {
				t.Fatalf("expected hostname: '%s', got '%s'", sandboxRequest.Config.Hostname, stdout)
			}
		})
	}
}
